# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import TypeAlias, TypedDict

from ..remotes import GitQuery


class PluginRegistryEntry(TypedDict):
    # fmt: off
    local:   None | str
    remote:  None | GitQuery
    # fmt: on


PluginRegistry: TypeAlias = dict[str, PluginRegistryEntry]
ZippedRegistry: TypeAlias = tuple[Path, PluginRegistry]


class ResolvedPluginRegistryEntry(TypedDict):
    # fmt: off
    local:   None | Path
    remote:  None | GitQuery
    # fmt: on


ResolvedPluginRegistry: TypeAlias = dict[str, ResolvedPluginRegistryEntry]
ResolvedZippedRegistry: TypeAlias = tuple[Path, ResolvedPluginRegistry]
